package org.dbxml.core.filer;

/*
 * dbXML License, Version 1.0
 *
 *
 * Copyright (c) 1999-2001 The dbXML Group, L.L.C.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        The dbXML Group (http://www.dbxml.com/)."
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 *
 * 4. The names "dbXML" and "The dbXML Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    info@dbxml.com.
 *
 * 5. Products derived from this software may not be called "dbXML",
 *    nor may "dbXML" appear in their name, without prior written
 *    permission of The dbXML Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE DBXML GROUP OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: MemFiler.java,v 1.12 2001/06/25 01:08:23 tom Exp $
 */

import org.dbxml.server.*;
import org.dbxml.core.*;
import org.dbxml.core.Collection;
import org.dbxml.core.data.*;

import java.util.*;

/**
 * MemFiler is an In-Memory Filer implementation for dbXML.  MemFiler can be
 * used for temporary collections and caching.  It's basically a layering on
 * top of HashMap.
 */

public final class MemFiler extends SimpleConfigurable implements Filer {
   private Map hashTable = null;
   private boolean opened = false;
   private boolean readOnly = false;
   private Collection collection;

   public MemFiler() {
      hashTable = Collections.synchronizedMap(new HashMap());
   }

   public MemFiler(Map hashTable, boolean readOnly) {
      this.hashTable = hashTable;
      this.readOnly = readOnly;
   }
   
   public MemFiler(Map hashTable) {
      this(hashTable, false);
   }
   
   public void setCollection(Collection collection) {
      this.collection = collection;
   }

   public String getName() {
      return "MemFiler";
   }

   private void checkOpened() throws DBException {
      if ( !opened )
         throw new FilerException(FaultCodes.COL_COLLECTION_CLOSED, "Filer is closed");
   }

   private void checkReadOnly() throws DBException {
      if ( readOnly )
         throw new FilerException(FaultCodes.COL_COLLECTION_READ_ONLY, "Filer is read-only");
   }
   
   /**
    * create creates a new DBObject and any associated resources for the new
    * DBObject, such as disk files, etc.
    *
    * @return Whether or not the DBObject was created
    */
   public boolean create() {
      hashTable.clear();
      return true;
   }

   /**
    * open opens the DBObject
    *
    * @return Whether or not the DBObject was opened
    */
   public boolean open() {
      opened = true;
      return opened;
   }

   /**
    * isOpened returns whether or not the DBObject is opened for business.
    *
    * @return The open status of the DBObject
    */
   public boolean isOpened() {
      return opened;
   }

   public boolean exists() {
      return true;
   }

   /**
    * drop instructs the DBObjectimplementation to remove itself from
    * existence.  The DBObject's parent is responsible for removing any
    * references to the DBObject in its own context.
    *
    * @return Whether or not the DBObject was dropped
    */
   public boolean drop() {
      hashTable.clear();
      opened = false;
      return !opened;
   }

   /**
    * close closes the DBObject
    *
    * @return Whether or not the DBObject was closed
    */
   public boolean close() {
      opened = false;
      return !opened;
   }
   
   /**
    * readRecord returns a Record from the Filer based on the specified
    * Key.
    *
    * @param key The Record's Key
    * @return The returned Record
    */
   public Record readRecord(Key key) throws DBException {
      checkOpened();
      return(Record)hashTable.get(key);
   }

   /**
    * writeRecord writes a Value to the Filer based on the specified Key.
    *
    * @param key The Record's Key
    * @param value The Record's Value
    * @return Whether or not the Record could be written
    */
   public boolean writeRecord(Key key, Value value) throws DBException {
      checkOpened();
      checkReadOnly();
      hashTable.put(key, new Record(key, value));
      return true;
   }

   /**
    * deleteRecord removes a Record from the Filer based on the
    * specified Key.
    *
    * @param key The Record's Key
    * @return Whether or not the Record was deleted
    */
   public boolean deleteRecord(Key key) throws DBException {
      checkOpened();
      checkReadOnly();
      return hashTable.remove(key) != null;
   }

   /**
    * getRecordCount returns the number of Records in the Filer.
    *
    * @return The Record count
    */
   public long getRecordCount() throws DBException {
      checkOpened();
      return hashTable.size();
   }

   /**
    * getRecordSet returns a RecordSet object for the current Filer.
    *
    * @return The Filer Enumerator
    */
   public RecordSet getRecordSet() throws DBException {
      checkOpened();
      return new MemRecordSet();
   }

   /**
    * MemRecordSet
    */
   private class MemRecordSet implements RecordSet {
      private Iterator enum = hashTable.values().iterator();

      public boolean hasMoreRecords() throws DBException {
         return enum.hasNext();
      }

      public Record getNextRecord() throws DBException {
         checkOpened();
         return(Record)enum.next();
      }

      public Value getNextValue() throws DBException {
         checkOpened();
         return((Record)enum.next()).getValue();
      }
      
      public Key getNextKey() {
         return((Record)enum.next()).getKey();
      }
   }
}


